package ru.landgrafhomyak.serdha.api.v0.ddl.module

import ru.landgrafhomyak.serdha.api.v0.ddl.table.Table
import ru.landgrafhomyak.serdha.api.v0.dml.Delete
import ru.landgrafhomyak.serdha.api.v0.dml.Insert
import ru.landgrafhomyak.serdha.api.v0.dml.InsertCreator
import ru.landgrafhomyak.serdha.api.v0.dml.Select
import ru.landgrafhomyak.serdha.api.v0.dml.Update
import ru.landgrafhomyak.serdha.api.v0.runtime.SynchronizedDatabase
import ru.landgrafhomyak.serdha.api.v0.runtime.Transaction

/**
 * Scope methods for creating new [module template][ModuleTemplate].
 *
 * @see ModuleTemplate.CreateModuleTemplatesScope.createTemplate
 * @see ModuleTemplate.ModuleCreator.createSchema
 */
public interface CreateModuleScope {
	/**
	 * Functional interface for creating table.
	 *
	 * Can be implemented like a companion object that just calls [TableUserExtension]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 *
	 * @see CreateModuleScope.createTable
	 * @see CreateModuleScope.createSessionScopeTemporaryTable
	 * @see CreateModuleScope.createTransactionScopeTemporaryTable
	 * @see CreateModuleScope.UpdateTable
	 */
	public interface TableCreator<TableUserExtension : Any> {
		/**
		 * Scope method that uses [creator] to initialize table schema.
		 *
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with descriptors for future access.
		 *
		 * @see CreateModuleScope.createTable
		 * @see CreateModuleScope.createSessionScopeTemporaryTable
		 * @see CreateModuleScope.createTransactionScopeTemporaryTable
		 * @see CreateModuleScope.UpdateTable
		 */
		public fun createTable(creator: ru.landgrafhomyak.serdha.api.v0.ddl.table.TableCreator<TableUserExtension>): TableUserExtension
	}

	/**
	 * Creates table in [specified namespace][namespace] with [specified name][name] and stores it in the database.
	 * If table was created by previous synchronizations, asserts that table at the specified path in the database is same as table provided by [initializer].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] this table must be [kept][CreateModuleScope.updateTable] (or [renamed][CreateModuleScope.renameTable]) or [deleted][CreateModuleScope.deleteTable].
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Creates a temporary table that exits only until connection to the database not [closed][SynchronizedDatabase._close].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createSessionScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Creates a temporary table that exits only inside [transaction][Transaction] and auto-deleted (or cleared, depends on driver implementation) when it [finished][Transaction._assertTransactionFinishedAndReleaseResources].
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type for containing table-related descriptors.
	 * @param namespace Way to group tables if there is a lot.
	 * @param name Name of table for debugging.
	 * @param initializer Table-related descriptors initializer.
	 * @return Table descriptor.
	 */
	public fun <TableUserExtension : Any> createTransactionScopeTemporaryTable(namespace: Namespace, name: String, initializer: TableCreator<TableUserExtension>): Table<TableUserExtension, Nothing>

	/**
	 * Functional interface for creating 'SELECT' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createSelectQuery
	 */
	public interface SelectQueryCreator<QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with updated descriptors for future access.
		 *
		 * @see CreateModuleScope.createSelectQuery
		 */
		public fun createSelect(creator: ru.landgrafhomyak.serdha.api.v0.dml.SelectCreator<QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'SELECT' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param QueryUserWrapper User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 */
	public fun <QueryUserWrapper : Any> createSelectQuery(initializer: SelectQueryCreator<QueryUserWrapper>): Select<QueryUserWrapper>

	/**
	 * Functional interface for creating 'INSERT ... VALUES' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createInsertParamsQuery
	 */
	public interface InsertParamsQueryCreator<TableUserExtension : Any, QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param table Descriptor of table to which values will be inserted.
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with updated descriptors for future access.
		 *
		 * @see CreateModuleScope.createInsertParamsQuery
		 */
		public fun createInsert(table: Table<TableUserExtension, *>, creator: InsertCreator.InsertParams<TableUserExtension, QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'INSERT ... VALUES' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserWrapper User's type for containing query-related descriptors.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createInsertParamsQuery(table: Table<TableUserExtension, *>, initializer: InsertParamsQueryCreator<TableUserExtension, QueryUserWrapper>): Insert.InsertParams<QueryUserWrapper>

	/**
	 * Functional interface for creating 'INSERT ... FROM' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createInsertFromQueryQuery
	 */
	public interface InsertFromQueryQueryCreator<TableUserExtension : Any, QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param table Descriptor of table to which values will be inserted.
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with updated descriptors for future access.
		 *
		 * @see CreateModuleScope.createInsertFromQueryQuery
		 */
		public fun createInsert(table: Table<TableUserExtension, *>, creator: InsertCreator.InsertFromQuery<TableUserExtension, QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'INSERT ... FROM' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createInsertFromQueryQuery(table: Table<TableUserExtension, *>, initializer: InsertFromQueryQueryCreator<TableUserExtension, QueryUserWrapper>): Insert.InsertFromQuery<QueryUserWrapper>

	/**
	 * Functional interface for creating 'UPDATE' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type with descriptors related to table, whose rows will be updated.
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createUpdateQuery
	 */
	public interface UpdateQueryCreator<TableUserExtension : Any, QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param table Descriptor of table which rows will be updated.
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with updated descriptors for future access.
		 *
		 * @see CreateModuleScope.createUpdateQuery
		 */
		public fun createUpdate(table: Table<TableUserExtension, *>, creator: ru.landgrafhomyak.serdha.api.v0.dml.UpdateCreator<TableUserExtension, QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'UPDATE' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table to which values will be inserted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createUpdateQuery(table: Table<TableUserExtension, *>, initializer: UpdateQueryCreator<TableUserExtension, QueryUserWrapper>): Update<QueryUserWrapper>

	/**
	 * Functional interface for creating 'DELETE' query.
	 *
	 * Can be implemented like a companion object that just calls [QueryUserWrapper]'s constructor with the same or similar signature.
	 *
	 * @param TableUserExtension User's type with descriptors related to table from which data will be deleted.
	 * @param QueryUserWrapper User's type with descriptors related to query.
	 *
	 * @see CreateModuleScope.createDeleteQuery
	 */
	public interface DeleteQueryCreator<TableUserExtension : Any, QueryUserWrapper : Any> {
		/**
		 * Scope method that uses [creator] to create a query.
		 *
		 * @param table Descriptor of table from which data will be deleted.
		 * @param creator Object with descriptor providers and factories.
		 * @return User's object with updated descriptors for future access.
		 *
		 * @see CreateModuleScope.createDeleteQuery
		 */
		public fun createDelete(table: Table<TableUserExtension, *>, creator: ru.landgrafhomyak.serdha.api.v0.dml.DeleteCreator<TableUserExtension, QueryUserWrapper>): QueryUserWrapper
	}

	/**
	 * Creates 'DELETE' query.
	 *
	 * On [module upgrading][ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate] must be recreated if needed.
	 *
	 * @param TableUserExtension User's type with descriptors related to table from which data will be deleted.
	 * @param initializer Query-related descriptors initializer.
	 * @return Query descriptor.
	 */
	public fun <TableUserExtension : Any, QueryUserWrapper : Any> createDeleteQuery(table: Table<TableUserExtension, *>, initializer: DeleteQueryCreator<TableUserExtension, QueryUserWrapper>): Delete<QueryUserWrapper>

	/**
	 * Replaces [specified namespace][rootNs] with schema provided by [template].
	 * [This namespace][rootNs] will be root namespace for module and shouldn't contain any other definitions.
	 *
	 * @param rootNs Namespace where the new module will be located.
	 * @param template Schema template.
	 * @return Module descriptor.
	 */
	public fun <ModuleUserExtension : Any> substituteModule(rootNs: Namespace, template: ModuleTemplate<ModuleUserExtension>): Module<ModuleUserExtension>
}