package ru.landgrafhomyak.serdha.api.v0.ddl.module

import ru.landgrafhomyak.serdha.api.v0.ddl.table.Table
import ru.landgrafhomyak.serdha.api.v0.ddl.table.TableCreator
import ru.landgrafhomyak.serdha.api.v0.ddl.table.TableUpdater
import ru.landgrafhomyak.serdha.api.v0.dml.Delete
import ru.landgrafhomyak.serdha.api.v0.dml.DeleteCreator
import ru.landgrafhomyak.serdha.api.v0.dml.Insert
import ru.landgrafhomyak.serdha.api.v0.dml.InsertCreator
import ru.landgrafhomyak.serdha.api.v0.dml.Select
import ru.landgrafhomyak.serdha.api.v0.dml.SelectCreator
import ru.landgrafhomyak.serdha.api.v0.dml.Update
import ru.landgrafhomyak.serdha.api.v0.dml.UpdateCreator
import ru.landgrafhomyak.serdha.api.v0.runtime.SynchronizedDatabase
import ru.landgrafhomyak.serdha.api.v0.runtime.Transaction

/**
 * Scope methods for upgrading [module template][ModuleTemplate].
 *
 * @see ModuleTemplate.CreateModuleTemplatesScope.upgradeTemplate
 * @see ModuleTemplate.ModuleUpdater.upgradeSchema
 */
public interface UpdateModuleScope : CreateModuleScope {
	/**
	 * Functional interface for updating table.
	 *
	 * Can be implemented like a companion object that just calls [TableNewUserExtension]'s constructor with the same or similar signature.
	 *
	 * @param TableNewUserExtension User's type with table's descriptors from the previous table version.
	 * @param TableOldUserExtension User's type with descriptors of updated table.
	 *
	 * @see UpdateModuleScope.updateTable
	 * @see UpdateModuleScope.renameTable
	 * @see UpdateModuleScope.updateAndRenameTable
	 * @see UpdateModuleScope.CreateTable
	 */
	public interface UpdateTable<TableNewUserExtension : Any, TableOldUserExtension : Any> {
		/**
		 * Scope method that uses [updater] to update table schema.
		 *
		 * @param oldTable Descriptor to the previous version of table from which can be got user's extension with old descriptors.
		 * @param updater Object with descriptor providers and factories.
		 * @return User's object with updated descriptors for future access.
		 *
		 * @see UpdateModuleScope.updateTable
		 * @see UpdateModuleScope.renameTable
		 * @see UpdateModuleScope.updateAndRenameTable
		 * @see UpdateModuleScope.CreateTable
		 */
		public fun updateTable(oldTable: Table<TableOldUserExtension, *>, updater: TableUpdater<TableNewUserExtension, TableOldUserExtension>): TableNewUserExtension
	}


	/**
	 * Marks that table wouldn't be changed in this module version.
	 *
	 * @param TableUserExtension User's type with descriptors of related to table.
	 * @param oldTable Descriptor of table created in a previous template version.
	 * @return New table descriptor.
	 */
	public fun <TableUserExtension : Any> keepTable(
		table: Table<TableUserExtension, *>,
	): Table<TableUserExtension, *>

	/**
	 * Renames or moves table without changing its schema.
	 *
	 * @param TableUserExtension User's type with descriptors of related to table.
	 * @param table Descriptor to table which should be renamed.
	 * @param newNamespace New namespace of the table. May be `null` if not changed.
	 * @param newName New namespace of the table. May be `null` if not changed.
	 * @return New table descriptor.
	 * @throws IllegalArgumentException If both [newName] and [newNamespace] are same as old or `null`.
	 */
	public fun <TableUserExtension : Any> renameTable(
		table: Table<TableUserExtension, *>,
		newNamespace: Namespace? = null,
		newName: String? = null
	): Table<TableUserExtension, *>

	/**
	 * Updates table's descriptors without changing its [name][Table.name] or [namespace][Table.namespacesFromModuleRoot].
	 *
	 * @param TableNewUserExtension User's type with descriptors of updated table.
	 * @param TableOldUserExtension User's type with table's descriptors from the previous table version.
	 * @param oldTable Descriptor to the previous version of table from which can be got user's extension with old descriptors.
	 * @param updater Object with descriptor providers and factories.
	 * @return New table descriptor.
	 */
	public fun <TableNewUserExtension : Any, TableOldUserExtension : Any> updateTable(
		oldTable: Table<TableOldUserExtension, *>,
		updater: UpdateTable<TableNewUserExtension, TableOldUserExtension>
	): Table<TableNewUserExtension, TableOldUserExtension>


	/**
	 * Both updates table's schema and renames or moves table.
	 *
	 * @param TableNewUserExtension User's type with descriptors of updated table.
	 * @param TableOldUserExtension User's type with table's descriptors from the previous table version.
	 * @param oldTable Descriptor of table which should be renamed and updated.
	 * @param newNamespace New namespace of the table. May be `null` if not changed.
	 * @param newName New namespace of the table. May be `null` if not changed.
	 * @param updater Object with descriptor providers and factories.
	 * @return New table descriptor.
	 * @throws IllegalArgumentException If both [newName] and [newNamespace] are same as old or `null`.
	 */
	public fun <TableNewUserExtension : Any, TableOldUserExtension : Any> updateAndRenameTable(
		oldTable: Table<TableNewUserExtension, TableOldUserExtension>,
		newNamespace: Namespace? = null,
		newName: String? = null,
		updater: UpdateTable<TableNewUserExtension, TableOldUserExtension>
	): Table<TableNewUserExtension, TableOldUserExtension>

	/**
	 * Deletes table and all data in it.
	 *
	 * @param table Descriptor of table which should be deleted.
	 */
	public fun deleteTable(table: Table<*, *>)

	/**
	 * Upgrades [module][oldModule] located at [rootNs] with new schema template.
	 *
	 * @param rootNs Namespace where the [old module][oldModule] is located and where upgaded module will be.
	 * @param template Schema template.
	 * @return Module descriptor.
	 */
	public fun <ModuleUserExtension : Any> upgradeModule(oldModule: Module<*>, rootNs: Namespace, template: ModuleTemplate<ModuleUserExtension>): Module<ModuleUserExtension>
}